/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.ui.rest.service.builds.buildsinfo.tabs.builddiff;

import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.build.ArtifactBuildAddon;
import org.artifactory.api.build.BuildService;
import org.artifactory.api.build.BuildWrapper;
import org.artifactory.api.build.model.diff.BuildsDiffBaseFileModel;
import org.artifactory.api.build.model.diff.BuildsDiffDependencyModel;
import org.artifactory.api.rest.build.diff.BuildsDiff;
import org.artifactory.build.BuildRun;
import org.artifactory.rest.common.service.ArtifactoryRestRequest;
import org.artifactory.rest.common.service.RestResponse;
import org.artifactory.rest.common.service.RestService;
import org.artifactory.rest.exception.ForbiddenException;
import org.artifactory.rest.exception.ForbiddenWebAppException;
import org.artifactory.ui.rest.model.builds.BuildDiffModel;
import org.artifactory.ui.rest.model.builds.BuildPropsModel;
import org.artifactory.ui.rest.model.builds.ModuleArtifactModel;
import org.artifactory.ui.rest.model.builds.ModuleDependencyModel;
import org.artifactory.ui.utils.DateUtils;
import org.jfrog.build.api.Build;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.config.BeanDefinition;
import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Component;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * @author Chen Keinan
 */
@Component
@Scope(BeanDefinition.SCOPE_PROTOTYPE)
public class BuildDiffService implements RestService {
    private static final Logger log = LoggerFactory.getLogger(BuildDiffService.class);

    private final BuildService buildService;
    private final AddonsManager addonsManager;

    @Autowired
    public BuildDiffService(BuildService buildService, AddonsManager addonsManager) {
        this.buildService = buildService;
        this.addonsManager = addonsManager;
    }

    @Override
    public void execute(ArtifactoryRestRequest request, RestResponse response) {
        try {
            String name = request.getPathParamByKey("name");
            String buildNumber = request.getPathParamByKey("number");
            String buildRepo = buildService.normaliseBuildRepoName(request.getQueryParamByKey("buildRepo"));
            String comparedBuildNum = request.getQueryParamByKey("otherNumber");
            String comparedDate = DateUtils.formatBuildDate(Long.parseLong(request.getQueryParamByKey("otherDate")));
            String buildStarted = DateUtils.formatBuildDate(Long.parseLong(request.getPathParamByKey("date")));

            // Exclude dependencies checkbox
            Boolean exDep = Boolean.valueOf(request.getQueryParamByKey("exDep"));

            buildService.assertReadPermissions(name, buildNumber, buildStarted, buildRepo);
            BuildDiffModel buildDiffModel = new BuildDiffModel();
            // fetch build diff data
            BuildRun buildRun = buildService.getBuildRun(name, buildNumber, buildStarted, buildRepo);
            if (!exDep) {
                BuildsDiff buildsDiff = fetchBuildsDiffData(name, comparedBuildNum, comparedDate, buildRepo, buildRun);
                List<ModuleArtifactModel> moduleArtifactModels = updateArtifact(buildsDiff);
                List<BuildPropsModel> buildPropsModels = updateProps(buildsDiff);
                List<ModuleDependencyModel> moduleDependencyModels = updateDependency(buildsDiff);
                updateDiffModel(moduleArtifactModels, buildPropsModels, moduleDependencyModels, buildDiffModel);
            } else {
                updateDependenciesWithExcludeInternal(name, comparedBuildNum, comparedDate, buildRepo, buildRun,
                        buildDiffModel);
            }
            // update build diff model
            response.iModel(buildDiffModel);
        } catch (ForbiddenWebAppException | ForbiddenException e) {
            throw e;
        } catch (Exception e) {
            log.error(e.toString());
            log.debug(e.toString(), e);
            response.error("error running build diff");
        }
    }

    /**
     * update dependencies with exclude internal
     *
     * @param name             - dependency name
     * @param comparedBuildNum - compared build num
     * @param comparedDate     - compared build date
     * @param buildRun         - build run
     * @param buildDiffModel   -- build diff model
     */
    private void updateDependenciesWithExcludeInternal(String name, String comparedBuildNum, String comparedDate,
            String comparedBuildRepo,
            BuildRun buildRun, BuildDiffModel buildDiffModel) {
        BuildRun secondBuildRun = buildService.getBuildRun(name, comparedBuildNum, comparedDate, comparedBuildRepo);
        ArtifactBuildAddon artifactBuildAddon = addonsManager.addonByType(ArtifactBuildAddon.class);
        BuildWrapper buildWrapper = buildService.getBuild(buildRun);
        BuildWrapper secondBuildWrapper = buildService.getBuild(secondBuildRun);
        if (buildWrapper == null || secondBuildWrapper == null) {
            return;
        }
        Build build = buildWrapper.getBuild();
        Build secondBuild = secondBuildWrapper.getBuild();
        Map<String, BuildsDiffBaseFileModel> artifactMap = new HashMap<>();
        List<BuildsDiffBaseFileModel> artifactDiff = artifactBuildAddon.compareArtifacts(build, secondBuild);
        artifactDiff.forEach(artifact -> artifactMap.put(artifact.getSha1(), artifact));
        List<BuildsDiffBaseFileModel> dependencyDiff = artifactBuildAddon.compareDependencies(build, secondBuild);
        List<ModuleDependencyModel> moduleDependencies = new ArrayList<>();
        dependencyDiff.forEach(dependency -> {
            if (artifactMap.get(dependency.getSha1()) == null) {
                moduleDependencies.add(new ModuleDependencyModel((BuildsDiffDependencyModel) dependency));
            }
        });
        buildDiffModel.setDependencies(moduleDependencies);
    }

    /**
     * update diff model
     *
     * @param moduleArtifactModels   - module artifact model
     * @param buildPropsModels       - builds props model
     * @param moduleDependencyModels - module dependency model
     * @param buildDiffModel         -  build diff model
     */
    private void updateDiffModel(List<ModuleArtifactModel> moduleArtifactModels, List<BuildPropsModel> buildPropsModels,
            List<ModuleDependencyModel> moduleDependencyModels, BuildDiffModel buildDiffModel) {
        buildDiffModel.setArtifacts(moduleArtifactModels);
        buildDiffModel.setDependencies(moduleDependencyModels);
        buildDiffModel.setProps(buildPropsModels);
    }

    /**
     * fetch build diff via comparator
     *
     * @param name             - build name
     * @param comparedBuildNum - compared build num
     * @param comparedDate     - compared build date
     * @param buildRun         - build run
     */
    private BuildsDiff fetchBuildsDiffData(String name,
            String comparedBuildNum, String comparedDate, String buildRepo, BuildRun buildRun) {
        BuildRun secondBuildRun = buildService.getBuildRun(name, comparedBuildNum, comparedDate, buildRepo);
        ArtifactBuildAddon artifactBuildAddon = addonsManager.addonByType(ArtifactBuildAddon.class);
        BuildWrapper build = buildService.getBuild(buildRun);
        BuildWrapper secondBuild = buildService.getBuild(secondBuildRun);
        if (build != null && secondBuild != null) {
            return artifactBuildAddon.getBuildsDiff(build.getBuild(), secondBuild.getBuild());
        }
        return null;
    }

    /**
     * update artifact diff list
     *
     * @param buildsDiff - build diff artifact list
     */
    private List<ModuleArtifactModel> updateArtifact(BuildsDiff buildsDiff) {
        List<ModuleArtifactModel> artifacts = new ArrayList<>();
        buildsDiff.artifacts.newItems.forEach(artifact -> artifacts.add(new ModuleArtifactModel(artifact)));
        buildsDiff.artifacts.removed.forEach(artifact -> artifacts.add(new ModuleArtifactModel(artifact)));
        buildsDiff.artifacts.unchanged.forEach(artifact -> artifacts.add(new ModuleArtifactModel(artifact)));
        buildsDiff.artifacts.updated.forEach(artifact -> artifacts.add(new ModuleArtifactModel(artifact)));
        return artifacts;
    }

    /**
     * update dependency diff list
     *
     * @param buildsDiff - build diff artifact list
     */
    private List<ModuleDependencyModel> updateDependency(BuildsDiff buildsDiff) {
        List<ModuleDependencyModel> dependencyList = new ArrayList<>();
        buildsDiff.dependencies.newItems
                .forEach(dependency -> dependencyList.add(new ModuleDependencyModel(dependency)));
        buildsDiff.dependencies.removed
                .forEach(dependency -> dependencyList.add(new ModuleDependencyModel(dependency)));
        buildsDiff.dependencies.unchanged
                .forEach(dependency -> dependencyList.add(new ModuleDependencyModel(dependency)));
        buildsDiff.dependencies.updated
                .forEach(dependency -> dependencyList.add(new ModuleDependencyModel(dependency)));
        return dependencyList;
    }

    /**
     * update props diff list
     *
     * @param buildsDiff - build diff artifact list
     */
    private List<BuildPropsModel> updateProps(BuildsDiff buildsDiff) {
        List<BuildPropsModel> props = new ArrayList<>();
        buildsDiff.properties.newItems.forEach(property -> props.add(new BuildPropsModel(property)));
        buildsDiff.properties.removed.forEach(property -> props.add(new BuildPropsModel(property)));
        buildsDiff.properties.unchanged.forEach(property -> props.add(new BuildPropsModel(property)));
        buildsDiff.properties.updated.forEach(property -> props.add(new BuildPropsModel(property)));
        return props;
    }
}
