package org.artifactory.ui.rest.service.admin.security.auth.currentuser;

import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.plugin.PluginsAddon;
import org.artifactory.api.config.CentralConfigService;
import org.artifactory.api.context.ArtifactoryContext;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.api.security.UserGroupService;
import org.artifactory.api.security.UserInfoBuilder;
import org.artifactory.descriptor.config.CentralConfigDescriptor;
import org.artifactory.descriptor.security.PasswordSettings;
import org.artifactory.descriptor.security.SecurityDescriptor;
import org.artifactory.rest.common.service.ArtifactoryRestRequest;
import org.artifactory.rest.common.service.RestResponse;
import org.artifactory.security.MutableUserInfo;
import org.artifactory.test.ArtifactoryHomeBoundTest;
import org.artifactory.ui.rest.model.admin.security.general.SecurityConfig;
import org.artifactory.ui.rest.model.admin.security.user.BaseUser;
import org.artifactory.ui.rest.service.admin.security.general.GetSecurityConfigService;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.MockitoAnnotations;
import org.springframework.test.util.ReflectionTestUtils;
import org.testng.annotations.AfterClass;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeClass;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.DataProvider;
import org.testng.annotations.Test;

import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.anyString;
import static org.mockito.Mockito.doNothing;
import static org.mockito.Mockito.when;
import static org.testng.Assert.assertFalse;
import static org.testng.Assert.assertTrue;

/**
 * @author Alexei Vainshtein
 */
@Test
public class GetCurrentUserServiceTest extends ArtifactoryHomeBoundTest {

    @Mock
    private ArtifactoryRestRequest requestMock;
    @Mock
    private RestResponse responseMock;
    @Mock
    private ArtifactoryContext artifactoryContextMock;
    @Mock
    private AddonsManager addonsManager;
    @Mock
    private PluginsAddon pluginsAddon;
    @Mock
    private CentralConfigService centralConfigService;
    @Mock
    private CentralConfigDescriptor configDescriptor;
    @Mock
    private AuthorizationService authorizationService;
    @Mock
    private SecurityDescriptor securityDescriptor;
    @Mock
    private GetSecurityConfigService securityConfigService;
    @Mock
    private UserGroupService userGroupService;

    private GetCurrentUserService service = new GetCurrentUserService();
    private MutableUserInfo userInfo;

    @BeforeClass
    public void setup() {
        MockitoAnnotations.initMocks(this);
        when(artifactoryContextMock.getCentralConfig()).thenReturn(centralConfigService);
        when(centralConfigService.getDescriptor()).thenReturn(configDescriptor);
        ReflectionTestUtils.setField(service, "addonsManager", addonsManager);
        ReflectionTestUtils.setField(service, "getSecurityConfigService", securityConfigService);
        ReflectionTestUtils.setField(service, "authService", authorizationService);
        ReflectionTestUtils.setField(service, "userGroupService", userGroupService);
        when(configDescriptor.getSecurity()).thenReturn(securityDescriptor);
        doNothing().when(securityConfigService).execute(any(), any());
        prepareAddonsManager();
        ArtifactoryContextThreadBinder.bind(artifactoryContextMock);
    }

    @BeforeMethod
    public void methodSetup() {
        userInfo = new UserInfoBuilder("myuser").build();
        when(userGroupService.currentUser()).thenReturn(userInfo);
    }

    @Test(description = "bugfix RTFACT-25151")
    public void testExecuteManageFlagsTrue() {
        userInfo.setPolicyManager(true);
        userInfo.setWatchManager(true);
        userInfo.setReportsManager(true);

        SecurityConfig securityConfig =
                new SecurityConfig(true, false, false, true, new PasswordSettings(), null, true, null);
        when(responseMock.getIModel()).thenReturn(securityConfig);
        prepareAuthorizationService();
        when(responseMock.iModel(any())).then(invocationOnMock -> {
            assertTrue(invocationOnMock.getArguments()[0] instanceof BaseUser);
            BaseUser userResponse = (BaseUser) invocationOnMock.getArguments()[0];
            assertTrue(userResponse.isWatchManager());
            assertTrue(userResponse.isPolicyManager());
            assertTrue(userResponse.isReportsManager());
            return null;
        });

        service.execute(requestMock, responseMock);
    }

    @Test(description = "bugfix RTFACT-25151")
    public void testExecuteManageFlagsDefault() {
        SecurityConfig securityConfig =
                new SecurityConfig(true, false, false, true, new PasswordSettings(), null, true, null);
        when(responseMock.getIModel()).thenReturn(securityConfig);
        prepareAuthorizationService();
        when(responseMock.iModel(any())).then(invocationOnMock -> {
            assertTrue(invocationOnMock.getArguments()[0] instanceof BaseUser);
            BaseUser userResponse = (BaseUser) invocationOnMock.getArguments()[0];
            assertFalse(userResponse.isWatchManager());
            assertFalse(userResponse.isPolicyManager());
            assertFalse(userResponse.isReportsManager());
            return null;
        });

        service.execute(requestMock, responseMock);
    }

    @Test(dataProvider = "archiveIndexEnabled", description = "RTFACT-21233")
    public void testExecute(boolean archiveIndexEnabled) {
        when(configDescriptor.isArchiveIndexEnabled()).thenReturn(archiveIndexEnabled);
        SecurityConfig securityConfig =
                new SecurityConfig(true, false, false, true, new PasswordSettings(), null, true, null);
        when(responseMock.getIModel()).thenReturn(securityConfig);
        prepareAuthorizationService();
        when(responseMock.iModel(any())).then(invocationOnMock -> {
            assertTrue(invocationOnMock.getArguments()[0] instanceof BaseUser);
            return null;
        });
        service.execute(requestMock, responseMock);
    }

    @AfterMethod
    public void resetMocks() {
        Mockito.reset(responseMock);
    }

    @AfterClass
    public void tearDown() {
        ArtifactoryContextThreadBinder.unbind();
    }

    private void prepareAuthorizationService() {
        when(authorizationService.isAdmin()).thenReturn(true);
        when(authorizationService.isUpdatableProfile()).thenReturn(true);
        when(authorizationService.requireProfileUnlock(anyString(), anyString())).thenReturn(false);
        when(authorizationService.requireProfilePassword()).thenReturn(false);
        when(authorizationService.isTransientUser()).thenReturn(true);
    }

    private void prepareAddonsManager() {
        when(addonsManager.addonByType(PluginsAddon.class)).thenReturn(pluginsAddon);
        when(addonsManager.isLicenseInstalled()).thenReturn(true);
        when(artifactoryContextMock.beanForType(AddonsManager.class)).thenReturn(addonsManager);
    }

    @DataProvider
    public static Object[][] archiveIndexEnabled() {
        return new Object[][] {
                {true},
                {false}
        };
    }
}